'''Command handling for Tarmac.'''
import httplib2
import logging
import os
import time

from bzrlib.commands import Command
from bzrlib.help import help_commands
from launchpadlib.launchpad import (Credentials, Launchpad, EDGE_SERVICE_ROOT,
    STAGING_SERVICE_ROOT)

from tarmac.bin import options
from tarmac.branch import Branch
from tarmac.hooks import tarmac_hooks
from tarmac.log import set_up_debug_logging, set_up_logging
from tarmac.exceptions import TarmacCommandError, TarmacMergeError
from tarmac.plugin import load_plugins
from tarmac.pidfile import get_pid, make_pidfile


class TarmacCommand(Command):
    '''A command class.'''

    NAME = None

    def __init__(self, registry):
        Command.__init__(self)

        self.config = registry.config
        self.registry = registry

        set_up_logging(self.config)
        self.logger = logging.getLogger('tarmac')

    def _usage(self):
        """Custom _usage for referencing "tarmac" instead of "bzr."""
        s = 'tarmac ' + self.name() + ' '
        for aname in self.takes_args:
            aname = aname.upper()
            if aname[-1] in ['$', '+']:
                aname = aname[:-1] + '...'
            elif aname[-1] == '?':
                aname = '[' + aname[:-1] + ']'
            elif aname[-1] == '*':
                aname = '[' + aname[:-1] + '...]'
            s += aname + ' '
        s = s[:-1]
        return s

    def run(self):
        '''Actually run the command.'''

    def get_launchpad_object(self, filename=None, staging=False):
        '''Return a Launchpad object for making API requests.'''
        # XXX: rockstar - 2009 Dec 13 - Ideally, we should be using
        # Launchpad.login_with, but currently, it doesn't support the option of
        # putting the credentials file somewhere other than where the cache
        # goes, and that's kinda nasty (and a security issue according to
        # Kees).
        if not filename:
            filename = self.config.CREDENTIALS

        if staging:
            SERVICE_ROOT = STAGING_SERVICE_ROOT
        else:
            SERVICE_ROOT = EDGE_SERVICE_ROOT

        self.logger.debug(
            "Connecting to the Launchpad API at {0}".format(SERVICE_ROOT))

        self.logger.debug("  Loading credentials from {0}".format(filename))
        if not os.path.exists(filename):
            self.logger.debug("  No existing API credentials were found")
            self.logger.debug("  Fetching new credentials from {0}".format(
                SERVICE_ROOT))

            launchpad = Launchpad.get_token_and_login(
                'Tarmac', SERVICE_ROOT, self.config.CACHE_HOME)
            launchpad.credentials.save(file(filename, 'w'))

            self.logger.debug("  Credentials saved to {0}".format(filename))
        else:
            credentials = Credentials()
            credentials.load(open(filename))
            self.logger.debug("  Credentials loaded".format(filename))

            launchpad = Launchpad(
                credentials, SERVICE_ROOT, self.config.CACHE_HOME)

        self.logger.debug("Connected")
        return launchpad


class cmd_authenticate(TarmacCommand):
    '''Create an OAuth token to be used by Tarmac.

    In order to use Tarmac at all, one must authenticate with Launchpad.  This
    command facilitates the process of getting an OAuth token from Launchpad.
    '''

    aliases = ['auth']
    takes_args = ['filename?']
    takes_options = [options.staging_option]

    def run(self, filename=None, staging=False):
        if os.path.exists(self.config.CREDENTIALS):
            self.logger.error('You have already been authenticated.')
        else:
            self.get_launchpad_object(filename=filename,
                                      staging=staging)


class cmd_help(TarmacCommand):
    '''Get help for Tarmac commands.'''

    aliases = ['fubar']
    takes_args = ['command?']

    def run(self, command=None):
        if command is None:
            self.outf.write('Usage:     tarmac <command> <options>\n\n')
            self.outf.write('Available commands:\n')
            self.help_commands()
        else:
            cmd = self.registry._get_command(None, command)
            if cmd is None:
                self.outf.write('Unknown command "%(command)s"\n' % {
                    'command': command})
                return
            text = cmd.get_help_text()
            if text:
                self.outf.write(text)

    def help_commands(self):
        help_commands(self.outf)


class cmd_merge(TarmacCommand):
    '''Automatically merge approved merge proposal branches.'''

    aliases = ['land']
    takes_args = ['branch_url?']
    takes_options = [
        options.http_debug_option,
        options.debug_option,
        options.imply_commit_message_option]

    def _do_merges(self, branch_url, imply_commit_message):

        target = Branch.create(
            self.launchpad.branches.getByUrl(url=branch_url), self.config,
            imply_commit_message=imply_commit_message)

        if not target.get_merge_proposals():
            return

        self.logger.debug('Firing tarmac_pre_merge hook')
        tarmac_hooks['tarmac_pre_merge'].fire(self, target)

        try:
            results = target.merge_branches()
            while True:
                try:
                    (hook, source, proposal) = results.next()
                    self.logger.debug('Firing %s hook' % hook)
                    tarmac_hooks[hook].fire(self, target, source, proposal)
                except StopIteration:
                    break
        # This except is here because we need the else and can't have it
        # without an except as well.
        except:
            raise
        else:
            self.logger.debug('Firing tarmac_post_merge hook')
            tarmac_hooks['tarmac_post_merge'].fire(self, target)
        finally:
            target.cleanup()

    def run(self, branch_url=None, debug=False, http_debug=False,
            launchpad=None, imply_commit_message=False):
        if debug:
            set_up_debug_logging()
            self.logger.debug('Debug logging enabled')
        if http_debug:
            httplib2.debuglevel = 1
            self.logger.debug('HTTP debugging enabled.')
        self.logger.debug('Loading plugins')
        load_plugins()
        self.logger.debug('Plugins loaded')

        self.launchpad = launchpad
        if self.launchpad is None:
            self.logger.debug('Loading launchpad object')
            self.launchpad = self.get_launchpad_object()
            self.logger.debug('launchpad object loaded')

        if branch_url:
            self.logger.debug(
                '%(branch_url)s specified as branch_url' % {
                    'branch_url': branch_url})
            if not branch_url.startswith('lp:'):
                raise TarmacCommandError('Branch urls must start with lp:')
            self._do_merges(branch_url, imply_commit_message)

        else:
            for branch in self.config.branches:
                self.logger.debug(
                    'Merging approved branches against %(branch)s' % {
                        'branch': branch})
                try:
                    self._do_merges(branch, imply_commit_message)
                except Exception, error:
                    self.logger.error(
                        'An error occurred trying to merge %s: %s',
                        branch, error)
                    raise


class cmd_bundle_merge(cmd_merge):
    """A command to merge multiple approved proposals.

    Bundle merge command merges all approved proposals into target. Once all
    approved proposals are merged the bundle_merge hook is fired. If the hook
    fails, the merge is reverted and proposals updated. If it succeeds,
    each proposal is committed individually into target.
    """

    def _do_merges(self, branch_url, imply_commit_message):
        existing_pid = get_pid()
        if existing_pid is None:
            make_pidfile()
        else:
            # Check how long the lock has been there.
            lock_creation_time = os.stat(self.config.PID_FILE)[-1]
            now = time.time()
            locked_period = now - lock_creation_time
            # If the locked_period is greater than the allowed, it probably 
            # means the test suite is stuck.
            message = 'Tarmac lock file in place for %d seconds, PID: %d'
            if locked_period > int(self.config.allowed_lock_period):
                self.logger.info(message % (locked_period, get_pid()))
                raise TarmacCommandError(
                    message % (locked_period, get_pid()))
            else:
                self.logger.info(message % (locked_period, get_pid()))
                return

        target = Branch.create(
            self.launchpad.branches.getByUrl(url=branch_url), self.config,
            imply_commit_message=imply_commit_message)

        if not target.get_merge_proposals():
            return

        self.logger.debug('Firing tarmac_pre_merge hook')
        tarmac_hooks['tarmac_pre_merge'].fire(self, target)

        try:
            # revno to restore to if bundle merge fails.
            restore_revno = target.tree.branch.revno() + 1
            results = target.merge_branches()
            collected_proposals = []
            while True:
                try:
                    (hook, source, proposal) = results.next()
                    collected_proposals.append((source, proposal))
                    self.logger.debug('Firing %s hook' % hook)
                    tarmac_hooks[hook].fire(self, target, source, proposal)
                except StopIteration:
                    break
            try:
                tarmac_hooks['tarmac_bundle_merge'].fire(
                    self, target, collected_proposals)
            except TarmacMergeError:
                # Uncommit and revert changes up to the point where target
                # was before the merge.
                target.uncommit_and_revert(restore_revno)

        # This except is here because we need the else and can't have it
        # without an except as well.
        except:
            raise
        else:
            self.logger.debug('Firing tarmac_post_merge hook')
            tarmac_hooks['tarmac_post_merge'].fire(self, target)
            # Push target.tree.branch to location branch_url.
            target.push()
        finally:
            target.cleanup()
