# Copyright 2009 Paul Hummer
# Copyright 2009 Canonical Ltd.
#
# This file is part of Tarmac.
#
# Authors: Paul Hummer
#          Rodney Dawes
#
# Tarmac is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by
# the Free Software Foundation.
#
# Tarmac is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tarmac.  If not, see <http://www.gnu.org/licenses/>.

'''Tarmac branch tools.'''
import logging
import os
import shutil
import tempfile

from bzrlib import branch as bzr_branch
from bzrlib.errors import NoSuchRevision, PointlessMerge
from bzrlib.uncommit import uncommit
from bzrlib.workingtree import WorkingTree

from tarmac.config import BranchConfig, TarmacConfig
from tarmac.exceptions import (
    BranchHasConflicts, TarmacMergeError, UnapprovedChanges)
from tarmac.utility import get_review_url


class Branch(object):

    def __init__(self, lp_branch, config=False, target=None,
                 imply_commit_message=False):
        self.lp_branch = lp_branch
        self.bzr_branch = bzr_branch.Branch.open(self.lp_branch.bzr_identity)
        self.config = BranchConfig(lp_branch.bzr_identity, config)
        self.imply_commit_message = imply_commit_message

        self.target = target
        self.logger = logging.getLogger('tarmac')


        # Determine whether or not we need to create a working tree.
        self.proposals = self.get_merge_proposals()
        if self.proposals:
            self.create_tree()
        else:
            self.logger.info(
                'No approved proposals found for %(branch_url)s' % {
                    'branch_url': self.lp_branch.bzr_identity})

    def __del__(self):
        """Do some potentially necessary cleanup during deletion."""
        try:
            # If we were using a temp directory, then remove it
            shutil.rmtree(self.temp_tree_dir)
        except AttributeError:
            # Not using a tempdir
            pass

    @classmethod
    def create(cls, lp_branch, config, target=None,
               imply_commit_message=False):
        return cls(lp_branch, config, target, imply_commit_message)

    def create_tree(self):
        '''Create the dir and working tree.'''
        try:
            self.logger.debug(
                'Using tree in %(tree_dir)s' % {
                    'tree_dir': self.config.tree_dir})
            if os.path.exists(self.config.tree_dir):
                self.tree = WorkingTree.open(self.config.tree_dir)
            else:
                self.logger.debug('Tree does not exist.  Creating dir')
                self.tree = self.bzr_branch.create_checkout(
                    self.config.tree_dir, lightweight=True)
        except AttributeError:
            # Store this so we can rmtree later
            self.temp_tree_dir = tempfile.mkdtemp()
            self.logger.debug(
                'Using temp dir at %(tree_dir)s' % {
                    'tree_dir': self.temp_tree_dir})
            self.tree = self.bzr_branch.create_checkout(self.temp_tree_dir)

        self.cleanup()

    def cleanup(self):
        '''Remove the working tree from the temp dir.'''
        assert self.tree
        self.tree.revert()
        for filename in [self.tree.abspath(f) for f in self.unmanaged_files]:
            if os.path.isdir(filename) and not os.path.islink(filename):
                shutil.rmtree(filename)
            else:
                os.remove(filename)

        self.tree.update()

    def merge(self, branch, revid=None):
        '''Merge from another tarmac.branch.Branch instance.'''
        assert self.tree
        conflict_list = self.tree.merge_from_branch(
            branch.bzr_branch, to_revision=revid)
        if conflict_list:
            message = u'Conflicts merging branch.'
            lp_comment = (
                u'Attempt to merge into %(target)s failed due to conflicts: '
                u'\n\n%(output)s' % {
                    'target': self.lp_branch.display_name,
                    "output": self.conflicts})
            raise BranchHasConflicts(message, lp_comment)

    def check_prerequisite(self, proposal):
        """Check if the given proposal prerequisite branch is merged."""
        prerequisite = proposal.prerequisite_branch
        if prerequisite:
            merges = [x for x in  prerequisite.landing_targets
                      if x.target_branch == self.lp_branch and
                      x.queue_status != u'Superseded']
            if len(merges) == 0:
                raise TarmacMergeError(
                    u'No proposals of prerequisite branch.',
                    u'No proposals found for merge of %s '
                    u'into %s.' % (
                        prerequisite.display_name,
                        self.lp_branch.display_name))
            elif len(merges) > 1:
                raise TarmacMergeError(
                    u'Too many proposals of prerequisite.',
                    u'More than one proposal found for merge '
                    u'of %s into %s, which is not Superseded.' % (
                        prerequisite.display_name,
                        self.lp_branch.display_name))
            elif len(merges) == 1:
                if merges[0].queue_status != u'Merged':
                    raise TarmacMergeError(
                        u'Prerequsiite not yet merged.',
                        u'The prerequisite %s has not yet been '
                        u'merged into %s.' % (
                            prerequisite.display_name,
                            self.lp_branch.display_name))

    def check_approved(self, source, proposal):
        """Check if the given source has no unapproved revisions."""
        approved = source.bzr_branch.revision_id_to_revno(
            str(proposal.reviewed_revid))
        tip = source.bzr_branch.revno()
        if tip > approved:
            message = u'Unapproved changes made after approval'
            lp_comment = (
                u'There are additional revisions which have not '
                u'been approved in review. Please seek review and '
                u'approval of these new revisions.')
            raise UnapprovedChanges(message, lp_comment)

    def merge_branches(self):
        """Perform the merge of approved branches.

        This function is a generator so control execution can be given back to
        the caller, so it can run hooks, and then execution resumed.
        """
        for proposal in self.get_merge_proposals():
            self.logger.debug(
                u'Preparing to merge %(source_branch)s' % {
                    'source_branch': proposal.source_branch.bzr_identity})
            source = Branch.create(
                proposal.source_branch, TarmacConfig(), target=self)

            try:
                self.check_prerequisite(proposal)

                if not proposal.reviewed_revid:
                    raise TarmacMergeError(
                        u'No approved revision specified.')

                self.check_approved(source, proposal)

                self.logger.debug(
                    'Merging %(source)s at revision %(revision)s' % {
                        'source': proposal.source_branch.display_name,
                        'revision': proposal.reviewed_revid})

                self.merge(
                    source,
                    str(proposal.reviewed_revid))
                # At this point we need to yield control back to the caller,
                # so the proper hook can be called.
                yield ("tarmac_pre_commit", source, proposal)

            except TarmacMergeError, failure:
                self.logger.warn(
                    u'Merging %(source)s into %(target)s failed: %(msg)s' %
                    {'source': proposal.source_branch.display_name,
                     'target': proposal.target_branch.display_name,
                     'msg': str(failure)})
                if failure.comment:
                    comment = failure.comment
                else:
                    comment = str(failure)
                self.update_proposal(proposal, comment)
                self.cleanup()
                continue
            except PointlessMerge:
                self.logger.warn(
                    'Merging %(source)s into %(target)s would be '
                    'pointless.' % {
                        'source': proposal.source_branch.display_name,
                        'target': proposal.target_branch.display_name})
                self.cleanup()
                continue

            merge_url = get_review_url(proposal)
            revprops = {'merge_url': merge_url}

            commit_message = proposal.commit_message
            if commit_message is None and self.imply_commit_message:
                commit_message = proposal.description
            self.commit(commit_message,
                         revprops=revprops,
                         authors=source.authors,
                         reviews=self._get_reviews(proposal))

            # At this point we need to yield control back to the caller,
            # so the proper hook can be called.
            yield ("tarmac_post_commit", source, proposal)
            self.cleanup()

    def update_proposal(self, proposal, comment):
        """Set proposal to Needs review and update with given comment."""
        subject = u'Re: [Merge] %(source)s into %(target)s' % {
            "source": proposal.source_branch.display_name,
            "target": proposal.target_branch.display_name}

        proposal.createComment(subject=subject,
                               content=comment)
        proposal.setStatus(status=u'Needs review')
        proposal.lp_save()

    def _get_reviews(self, proposal):
        """Get the set of reviews from the proposal."""
        reviews = []
        for vote in proposal.votes:
            if not vote.comment:
                continue
            else:
                reviews.append('%s;%s' % (vote.reviewer.display_name,
                                          vote.comment.vote))

        if len(reviews) == 0:
            return None

        return reviews

    @property
    def unmanaged_files(self):
        """Get the list of ignored and unknown files in the tree."""
        self.tree.lock_read()
        unmanaged = [x for x in self.tree.unknowns()]
        unmanaged.extend([x[0] for x in self.tree.ignored_files()])
        self.tree.unlock()
        return unmanaged

    @property
    def conflicts(self):
        '''Print the conflicts.'''
        assert self.tree.conflicts()
        conflicts = []
        for conflict in self.tree.conflicts():
            conflicts.append(
                u'%s in %s' % (conflict.typestring, conflict.path))
        return '\n'.join(conflicts)

    def commit(self, commit_message, revprops=None, **kwargs):
        '''Commit changes.'''
        if not revprops:
            revprops = {}

        authors = kwargs.pop('authors', None)
        reviews = kwargs.pop('reviews', None)

        if not authors:
            authors = self.authors

        if reviews:
            for review in reviews:
                if '\n' in review:
                    raise TarmacMergeError('\\n is not a valid character in a '
                                           'review identity or vote.')
            revprops['reviews'] = '\n'.join(reviews)

        self.tree.commit(commit_message, committer='Tarmac',
                         revprops=revprops, authors=authors)

    @property
    def landing_candidates(self):
        '''Wrap the LP representation of landing_candidates.'''
        return self.lp_branch.landing_candidates

    @property
    def authors(self):
        author_list = []

        if self.target:
            self.bzr_branch.lock_read()
            self.target.bzr_branch.lock_read()

            graph = self.bzr_branch.repository.get_graph(
                self.target.bzr_branch.repository)

            unique_ids = graph.find_unique_ancestors(
                self.bzr_branch.last_revision(),
                [self.target.bzr_branch.last_revision()])

            revs = self.bzr_branch.repository.get_revisions(unique_ids)
            for rev in revs:
                apparent_authors = rev.get_apparent_authors()
                for author in apparent_authors:
                    author.replace('\n', '')
                    if author not in author_list:
                        author_list.append(author)

            self.target.bzr_branch.unlock()
            self.bzr_branch.unlock()
        else:
            last_rev = self.bzr_branch.last_revision()
            if last_rev != 'null:':
                rev = self.bzr_branch.repository.get_revision(last_rev)
                apparent_authors = rev.get_apparent_authors()
                author_list.extend(
                    [a.replace('\n', '') for a in apparent_authors])

        return author_list

    def uncommit_and_revert(self, revno):
        """Uncommit and revert changes up to revno."""
        uncommit(self.tree.branch, revno=revno)
        self.tree.revert()

    def push(self):
        """Push changes to the remote branch."""
        self.tree.branch.push(self.bzr_branch, overwrite=True)

    @property
    def fixed_bugs(self):
        """Return the list of bugs fixed by the branch."""
        bugs_list = []

        self.bzr_branch.lock_read()
        oldrevid = self.bzr_branch.get_rev_id(self.lp_branch.revision_count)
        for rev_info in self.bzr_branch.iter_merge_sorted_revisions(
            stop_revision_id=oldrevid):
            try:
                rev = self.bzr_branch.repository.get_revision(rev_info[0])
                for bug in rev.iter_bugs():
                    if bug[0].startswith('https://launchpad.net/bugs/'):
                        bugs_list.append(bug[0].replace(
                                'https://launchpad.net/bugs/', ''))
            except NoSuchRevision:
                continue

        self.bzr_branch.unlock()
        return bugs_list

    def get_merge_proposals(self):
        """Get a target branch's merge proposals."""
        try:
            return self._proposals
        except AttributeError:
            pass

        self._proposals = []
        for entry in self.landing_candidates:
            self.logger.debug("Considering merge proposal: {0}".format(entry))

            if entry.queue_status != u'Approved':
                self.logger.debug(
                    "  Skipping proposal: status is {0}, not "
                    "'Approved'".format(entry.queue_status))
                continue

            if not self.imply_commit_message and not entry.commit_message:
                self.logger.debug(
                    "  Skipping proposal: proposal has no commit message")
                continue

            self._proposals.append(entry)

        return self._proposals


