# Copyright 2009 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

# Code from lib/canonical/lazr/pidfile.py modified to be used here.

__metaclass__ = type

import tempfile
import os
import atexit
import sys
from signal import signal, SIGTERM

from tarmac.config import TarmacConfig


def make_pidfile():
    """Write the current process id to a PID file.

    Also installs an atexit handler to remove the file on process termination.

    Also installs a SIGTERM signal handler to remove the file on SIGTERM.
    If you install your own handler, you will want to call remove_pidfile
    inside it.
    """
    pidfile = TarmacConfig().PID_FILE
    if os.path.exists(pidfile):
        raise RuntimeError("PID file %s already exists. Already running?" %
                pidfile)

    atexit.register(remove_pidfile)
    def remove_pidfile_handler(*ignored):
        sys.exit(-1 * SIGTERM)
    signal(SIGTERM, remove_pidfile_handler)

    fd, tempname = tempfile.mkstemp(dir=os.path.dirname(pidfile))
    outf = os.fdopen(fd, 'w')
    outf.write(str(os.getpid())+'\n')
    outf.flush()
    outf.close()
    os.rename(tempname, pidfile)


def remove_pidfile():
    """Remove the PID file.

    This should only be needed if you are overriding the default SIGTERM
    signal handler.
    """
    pidfile = TarmacConfig().PID_FILE
    pid = get_pid()
    if pid is None:
        return
    if pid == os.getpid():
        os.unlink(pidfile)


def get_pid():
    """Return the PID as an integer, or None

    May raise a ValueError if the PID file is corrupt.

    This method will only be needed by service or monitoring scripts.

    Currently no checking is done to ensure that the process is actually
    running, is healthy, or died horribly a while ago and its PID being
    used by something else. What we have is probably good enough.
    """
    pidfile = TarmacConfig().PID_FILE
    try:
        pid = open(pidfile).read()
        return int(pid)
    except IOError:
        return None
    except ValueError:
        raise ValueError("Invalid PID %s" % repr(pid))

