# Copyright 2009 Paul Hummer
# This file is part of Tarmac.
#
# Tarmac is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by
# the Free Software Foundation.
#
# Tarmac is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tarmac.  If not, see <http://www.gnu.org/licenses/>.

'''Tarmac plugin for running tests on multiple proposals.'''

# Head off lint warnings.
os = None
subprocess = None

import datetime
from email import MIMEMultipart, MIMEText
from email.mime.application import MIMEApplication
import gzip
from StringIO import StringIO
import tempfile
import unittest

import bzrlib.config

from bzrlib.lazy_import import lazy_import
from bzrlib.smtp_connection import SMTPConnection
lazy_import(globals(), '''
    import os
    import subprocess
    ''')

import subunit

from tarmac.exceptions import TarmacMergeError
from tarmac.hooks import TarmacHookPoint, tarmac_hooks
from tarmac.plugins import TarmacPlugin

_double_line = '=' * 70
_single_line = '-' * 70

def format_error(flavor, test, error):
    return '\n'.join(
        [_double_line,
         '%s: %s' % (flavor, test),
         _single_line,
         error,
         ''])


class TestResultFormatter:

    def __init__(self, start_time, end_time, result, proposal):
        self.result = result
        self.start_time = start_time
        self.source_branch = proposal.source_branch.display_name
        self.source_revno = proposal.source_branch.revision_count
        self.target_branch = proposal.target_branch.display_name
        self.target_revno = proposal.target_branch.revision_count
        self.num_tests = self.result.testsRun
        self.duration = end_time - start_time
        self.num_failures = len(self.result.failures)
        self.num_errors = len(self.result.errors)

    def _format_test_list(self, header, tests):
        if not tests:
            return []
        tests = ['  ' + test.id() for test, error in tests]
        return [header, '-' * len(header)] + tests + ['']

    def get_full_result(self):
        return '\n'.join(self.full_result)

    def format_result(self, attached_log=False):
        output = [
            'Tests started at approximately %s' % self.start_time,
            ]
        output.append('Source: %s r%s' % (
            self.source_branch, self.source_revno))
        output.append('Target: %s r%s' % (
            self.target_branch, self.target_revno))
        output.extend([
            '',
            '%s tests run in %s, %s failures, %s errors' % (
                self.num_tests, self.duration, self.num_failures,
                self.num_errors),
            '',
            ])

        bad_tests = (
            self._format_test_list('Failing tests', self.result.failures) +
            self._format_test_list('Tests with errors', self.result.errors))
        output.extend(bad_tests)

        if bad_tests:
            email_summary = StringIO()
            for test, error in self.result.failures:
                email_summary.write(
                    format_error('FAILURE', test, error))
            for test, error in self.result.errors:
                email_summary.write(
                    format_error('ERROR', test, error))
            output.append(email_summary.getvalue())

        # Include notification about the attached log. Useful when the output
        # is sent as an email.
        if attached_log:
            output.extend(['(See the attached file for the complete log)', ''])
        return '\n'.join(output)


class VerifyCommandFailed(TarmacMergeError):
    """Running the verify_command failed."""


class BundleCommand(TarmacPlugin):

    def __init__(self, smtp_connection=None):
        super(BundleCommand, self).__init__()
        if smtp_connection is None:
            smtp_connection = SMTPConnection(bzrlib.config.GlobalConfig())
        self._smtp_connection = smtp_connection

    def run(self, command, target, collected_proposals):
        """Plugin for running the registered bundle verify command on target.

        This is very similar to the Command plugin, but instead of running the
        verify command on target for individual proposals, it gets a list of
        proposals merged into target and then runs the verify command.
        Updating the list of proposals in case of failure.
        """
        try:
            self.verify_command = target.config.bundle_verify_command
        except AttributeError:
            # This can be killed two versions after 0.4, whatever version that
            # is.
            try:
                self.verify_command = target.config.test_command
                self.logger.warn(
                    'test_command config setting is deprecated. '
                    'Please use verify_command instead.')
            except AttributeError:
                return

        start_time = datetime.datetime.utcnow()
        self.logger.debug('Running test command: %s' % self.verify_command)
        cwd = os.getcwd()
        os.chdir(target.tree.abspath(''))
        output = tempfile.TemporaryFile()
        try:
            proc = subprocess.Popen(
                self.verify_command,
                shell=True,
                stdout=output,
                stderr=subprocess.STDOUT)
            return_code = proc.wait()
            output.seek(0)
            os.chdir(cwd)
            self.logger.debug('Completed test command: %s' % self.verify_command)
            end_time = datetime.datetime.utcnow()

            if return_code != 0:
                for source, proposal in collected_proposals:
                    self.do_failed(
                        start_time, end_time, proposal, output)
                message = u'Test command "%s" failed.' % self.verify_command
                comment = (
                    u'Bundle lander failed to land proposals: %s' %
                    collected_proposals)
                raise VerifyCommandFailed(message, comment)
            else:
                self.logger.info('==========')
                self.logger.info(output.read())
        finally:
            output.close()

    def do_failed(self, start_time, end_time, proposal, output):
        '''Perform failure tests.

        In this case, the full failure is emailed to the proposal subscribers,
        a summary of the test failure is posted as a comment and the
        proposal is then set to "Needs Review" so that Tarmac doesn't attempt
        to merge it again without human interaction.
        '''
        # Make output into a Result object, and create a tgz of the
        # full log.
        # We will use these for the gzip.
        fd, path = tempfile.mkstemp()
        os.close(fd)
        gz = gzip.open(path, 'wb')
        # These are for creating the test result object.
        result = unittest.TestResult()
        discard = tempfile.TemporaryFile()
        # The temporary file gets things that subunit doesn't understand.
        # It prints them to stdout by default.  We don't want the output on
        # stdout, so we get it and discard it.  (Note that the things subunit
        # doesn't understand are still put in the full_result, so that the
        # email will include them.)
        subunit_server = subunit.TestProtocolServer(result, discard)
        for line in output:
            subunit_server.lineReceived(line)
            gz.write(line)
        gz.close()
        try:
            gzipped_log = open(path).read()
        finally:
            os.unlink(path)
        formatter = TestResultFormatter(
            start_time, end_time, result, proposal)
        output.seek(0)

        emails = self.get_subscribers_emails(proposal)
        self.send_report_email(emails, formatter, gzipped_log)

        summary = formatter.format_result()
        message = u'Test command "%s" failed.' % self.verify_command
        comment = (u'The attempt to merge %(source)s into %(target)s failed. '
                   u'Below is a summary from the failed tests.\n\n'
                   u'%(summary)s') % {
            'source': proposal.source_branch.display_name,
            'target': proposal.target_branch.display_name,
            'summary': summary,
            }

        self.logger.warn(
            u'Merging %(source)s into %(target)s failed: %(msg)s' %
            {'source': proposal.source_branch.display_name,
             'target': proposal.target_branch.display_name,
             'msg': str(message)})

        subject = u'Re: [Merge] %(source)s into %(target)s' % {
            "source": proposal.source_branch.display_name,
            "target": proposal.target_branch.display_name}

        proposal.createComment(subject=subject,
                               content=comment)
        proposal.setStatus(status=u'Needs Review')
        proposal.lp_save()

    def get_subscribers_emails(self, proposal):
        """Return a list of the proposal subscribers emails."""
        subscribers_emails = set()
        # Get all the reviewers.
        for vote in proposal.votes:
            subscribers_emails.add(
                vote.reviewer.preferred_email_address.email)
        # Get source branch subscribers.
        for subscriber in proposal.source_branch.subscribers:
            subscribers_emails.add(subscriber.preferred_email_address.email)
        return subscribers_emails

    def send_report_email(self, emails, formatter, full_log_gz):
        """Send an email summarizing the test results.

        :param emails: The list of emails to mail to.
        :param formatter: The TestResultFormatter object.
        :param full_test_result: The full log.
        """
        message = self._build_report_email(emails, formatter, full_log_gz)
        self._send_email(message)

    def _build_report_email(self, emails, formatter, full_log_gz):
        """Build a MIME email summarizing the test results.

        :param emails: The list of emails to mail to.
        :param formatter: The TestResultFormatter object.
        :param full_log_gz: A gzip of the full log.
        """
        message = MIMEMultipart.MIMEMultipart()
        message['To'] = ', '.join(emails)
        message['From'] = "Tarmac Bundle lander."
        subject = 'Test results: %s => %s: %s' % (
            formatter.source_branch, formatter.target_branch, "FAILURE")
        message['Subject'] = subject

        body_text = formatter.format_result(attached_log=True)
        # Make the body.
        body = MIMEText.MIMEText(body_text, 'plain', 'utf8')
        body['Content-Disposition'] = 'inline'
        message.attach(body)

        # Attach the gzipped log.
        zipped_log = MIMEApplication(full_log_gz, 'x-gzip')
        zipped_log.add_header(
            'Content-Disposition', 'attachment',
            filename='%s-r%s.subunit.gz' % (
                formatter.source_branch, formatter.source_revno))
        message.attach(zipped_log)
        return message

    def _send_email(self, message):
        """Actually send 'message'."""
        self._smtp_connection.send_email(message)

# Create the hook here so we don't change the tarmac_hooks API with this very
# specific hook.
tarmac_hooks.create_hook(TarmacHookPoint('tarmac_bundle_merge',
    'Called by the bundle merge command.',
    (0, 3, 4), False))
tarmac_hooks['tarmac_bundle_merge'].hook(
    BundleCommand(), 'BundleCommand plugin')
