"""Tests for the BundleCommmand plug-in."""

from datetime import datetime, timedelta
import doctest
import os
import unittest

from testtools import TestCase, TestResult
from testtools.matchers import DocTestMatches

from tarmac.bin.registry import CommandRegistry
from tarmac.plugins.bundlecommand import (
    BundleCommand, TestResultFormatter, VerifyCommandFailed)
from tarmac.tests import TarmacTestCase
from tarmac.tests.test_commands import FakeCommand
from tarmac.tests.mock import Thing, MockMergeProposal


class LoggingSMTPConnection(object):
    """An SMTPConnection double that logs sent email."""

    def __init__(self, log):
        self._log = log

    def send_email(self, message):
        self._log.append(message)


class Helpers:

    def make_formatter(self):
        proposal = Thing(
            source_branch=Thing(
                revision_count=123,
                display_name="lp:project/source"),
            target_branch=Thing(
                revision_count=120,
                display_name="lp:project"))
        start_time = datetime.utcnow()
        end_time = start_time + timedelta(hours=1)

        class SomeTest(TestCase):
            def test_ok(self):
                pass
            def test_fail(self):
                self.fail("oh no")
            def test_error(self):
                1/0
        self.fail_test = SomeTest('test_fail')
        self.error_test = SomeTest('test_error')
        test = unittest.TestSuite(
            [self.fail_test, self.error_test, SomeTest('test_ok')])
        result = TestResult()
        test.run(result)

        return TestResultFormatter(start_time, end_time, result, proposal)

    def lp_save(self, *args, **kwargs):
        """Dummy lp_save method."""
        pass


class TestTestResultFormatter(TestCase, Helpers):
    """Tests for `TestResultFormatter`."""

    expected_failure_summary_email = ("""\
Tests started at approximately %(start_time)s
Source: %(source_branch)s r%(source_revno)s
Target: %(target_branch)s r%(target_revno)s
<BLANKLINE>
%(num_tests)s tests run in %(duration)s, %(num_failures)s failures, %(num_errors)s errors
<BLANKLINE>
Failing tests
-------------
  %(fail_id)s
<BLANKLINE>
Tests with errors
-----------------
  %(error_id)s
<BLANKLINE>
======================================================================
FAILURE: test_fail...
----------------------------------------------------------------------
Text attachment: traceback
------------
Traceback (most recent call last):
...
------------
<BLANKLINE>
======================================================================
ERROR: test_error...
----------------------------------------------------------------------
Text attachment: traceback
------------
Traceback (most recent call last):
...
------------
<BLANKLINE>
<BLANKLINE>
(See the attached file for the complete log)
""")

    expected_failure_summary = ("""\
Tests started at approximately %(start_time)s
Source: %(source_branch)s r%(source_revno)s
Target: %(target_branch)s r%(target_revno)s
<BLANKLINE>
%(num_tests)s tests run in %(duration)s, %(num_failures)s failures, %(num_errors)s errors
<BLANKLINE>
Failing tests
-------------
  %(fail_id)s
<BLANKLINE>
Tests with errors
-----------------
  %(error_id)s
<BLANKLINE>
======================================================================
FAILURE: test_fail...
----------------------------------------------------------------------
Text attachment: traceback
------------
Traceback (most recent call last):
...
------------
<BLANKLINE>
======================================================================
ERROR: test_error...
----------------------------------------------------------------------
Text attachment: traceback
------------
Traceback (most recent call last):
...
------------
<BLANKLINE>
""")

    def test_format_result_with_attached_log(self):
        formatter = self.make_formatter()
        summary = formatter.format_result(attached_log=True)
        data = {
            'start_time': formatter.start_time,
            'source_branch': formatter.source_branch,
            'source_revno': formatter.source_revno,
            'target_branch': formatter.target_branch,
            'target_revno': formatter.target_revno,
            'num_tests': formatter.num_tests,
            'duration': formatter.duration,
            'num_failures': formatter.num_failures,
            'num_errors': formatter.num_errors,
            'fail_id': self.fail_test.id(),
            'error_id': self.error_test.id()
            }
        self.assertThat(summary, DocTestMatches(
            self.expected_failure_summary_email % data, doctest.REPORT_NDIFF|
            doctest.ELLIPSIS))

    def test_format_result_without_attached_log(self):
        formatter = self.make_formatter()
        summary = formatter.format_result(attached_log=False)
        data = {
            'start_time': formatter.start_time,
            'source_branch': formatter.source_branch,
            'source_revno': formatter.source_revno,
            'target_branch': formatter.target_branch,
            'target_revno': formatter.target_revno,
            'num_tests': formatter.num_tests,
            'duration': formatter.duration,
            'num_failures': formatter.num_failures,
            'num_errors': formatter.num_errors,
            'fail_id': self.fail_test.id(),
            'error_id': self.error_test.id()
            }
        self.assertThat(summary, DocTestMatches(
            self.expected_failure_summary % data, doctest.REPORT_NDIFF|
            doctest.ELLIPSIS))


class BundleCommandTests(TarmacTestCase, Helpers):
    """Test the BundleCommand."""

    def setUp(self):
        """Set up additional data we need for tests."""
        super(BundleCommandTests, self).setUp()
        self.collected_proposals = [
            (Thing(),
            MockMergeProposal(
                Thing(
                    display_name='lp:project/source',
                    revision_count=123,
                    subscribers=[
                        Thing(preferred_email_address=Thing(
                            email='foo@example.com'))]),
                Thing(
                    display_name='lp:project',
                    revision_count=120),
                status="Approved",
                votes=[
                    Thing(reviewer=Thing(
                        preferred_email_address=Thing(
                        email='foo.bar@example.com')))])),
            (Thing(),
            MockMergeProposal(
                Thing(
                    display_name='lp:project/source2',
                    revision_count=124,
                    subscribers=[
                        Thing(preferred_email_address=Thing(
                            email='foo@example.com'))]),
                Thing(
                    display_name='lp:project',
                    revision_count=120),
                status="Approved",
                votes=[
                    Thing(reviewer=Thing(
                        preferred_email_address=Thing(
                        email='foo.bar@example.com')))])),
            ]
        # Patch the plugin smtp_connection so emails are never actually sent.
        email_log = []
        smtp_connection = LoggingSMTPConnection(email_log)
        self.plugin = BundleCommand(smtp_connection=smtp_connection)
        registry = CommandRegistry(config=self.config)
        self.command = FakeCommand(registry)

    def test_run(self):
        """Test that the plug-in run without errors."""
        target = Thing(config=Thing(
            bundle_verify_command="/bin/true"),
            tree=Thing(abspath=os.path.abspath))
        self.plugin.run(
            command=self.command, target=target,
            collected_proposals=self.collected_proposals)

    def test_run_failure(self):
        """Test that a failure raises the correct exception."""
        # No email in the log.
        self.assertEqual(self.plugin._smtp_connection._log, [])
        # Proposal status are approved.
        proposal1, proposal2 = self.collected_proposals
        self.assertEqual(proposal1[1].queue_status, "Approved")
        self.assertEqual(proposal2[1].queue_status, "Approved")

        target = Thing(config=Thing(
            bundle_verify_command="/bin/false"),
            tree=Thing(abspath=os.path.abspath))
        self.assertRaises(VerifyCommandFailed,
            self.plugin.run,
            command=self.command, target=target,
            collected_proposals=self.collected_proposals)

        # There are two emails in the log. One for each proposal.
        email = self.plugin._smtp_connection._log.pop()
        self.assertEqual(
            email.get('Subject'),
            'Test results: lp:project/source2 => lp:project: FAILURE')
        email = self.plugin._smtp_connection._log.pop()
        self.assertEqual(
            email.get('Subject'),
            'Test results: lp:project/source => lp:project: FAILURE')

        # Proposals status is updated.
        proposal1, proposal2 = self.collected_proposals
        self.assertEqual(proposal1[1].queue_status, "Needs Review")
        self.assertEqual(proposal2[1].queue_status, "Needs Review")

    def test_send_report_email(self):
        """Test send_report_email()."""
        formatter = self.make_formatter()

        emails = self.plugin.get_subscribers_emails(
            self.collected_proposals[0][1])
        self.plugin.send_report_email(
            emails, formatter, 'some gzip content')
        email = self.plugin._smtp_connection._log.pop()
        self.assertEqual(
            email.get('To'), 'foo@example.com, foo.bar@example.com')
        self.assertEqual(
            email.get('Subject'),
            "Test results: lp:project/source => lp:project: FAILURE")
