# Copyright 2009 Paul Hummer
# Copyright 2009 Canonical Ltd.
#
# This file is part of Tarmac.
#
# Tarmac is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by
# the Free Software Foundation.
#
# Tarmac is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tarmac.  If not, see <http://www.gnu.org/licenses/>.

'''Tests for tarmac.branch'''
import os
import shutil

from bzrlib.errors import PointlessMerge

from tarmac import branch
from tarmac.tests import BranchTestCase
from tarmac.tests.mock import MockLPBranch, Thing


class TestBranch(BranchTestCase):
    '''Test for Tarmac.branch.Branch.'''

    def test_create(self):
        '''Test the creation of a TarmacBranch instance.'''
        mock = self.make_branch()
        self.add_branch_config(mock.tree_dir)

        a_branch = branch.Branch.create(mock, self.config)
        self.assertTrue(isinstance(a_branch, branch.Branch))
        self.assertTrue(a_branch.lp_branch.bzr_identity is not None)
        self.assertFalse(hasattr(a_branch, 'tree'))
        self.remove_branch_config(mock.tree_dir)

    def test_create_with_tree(self):
        '''Test the creation of a TarmacBranch with a created tree.'''
        self.assertTrue(isinstance(self.branch1, branch.Branch))
        self.assertTrue(self.branch1.lp_branch.bzr_identity is not None)
        self.assertTrue(hasattr(self.branch1, 'tree'))

    def test_merge_raises_exception_with_no_tree(self):
        '''A merge on a branch with no tree will raise an exception.'''
        mock = self.make_branch(source_branch=self.branch1.lp_branch)
        self.add_branch_config(mock.tree_dir)
        branch3 = branch.Branch.create(mock, self.config)

        self.assertRaises(
            AttributeError, branch3.merge, self.branch2)
        self.remove_branch_config(mock.tree_dir)

    def test_merge_no_changes(self):
        '''A merge on a branch with a tree will raise an exception if no
        changes are present.'''
        mock = self.make_branch(source_branch=self.branch1.lp_branch)
        self.add_branch_config(mock.tree_dir)
        branch3 = branch.Branch.create(mock, self.config)

        self.assertRaises(PointlessMerge, self.branch1.merge, branch3)
        self.remove_branch_config(mock.tree_dir)

    def test_merge(self):
        '''A merge on a branch with a tree of a branch with changes will merge.
        '''
        self.branch1.merge(self.branch2)
        self.assertTrue(self.branch1.tree.changes_from(
                self.branch1.tree.basis_tree()).has_changed())

    def test_merge_with_authors(self):
        '''A merge from a branch with authors'''
        authors = ['author1', 'author2']
        self.branch2.commit('Authors test', authors=authors)
        self.branch2.commit('Another author', authors=['author3'])
        self.branch2.lp_branch.revision_count += 2
        orig_authors = self.branch2.authors
        self.branch1.merge(self.branch2)
        self.branch1.commit('Authors Merge test', authors=self.branch2.authors)
        self.branch1.lp_branch.revision_count += 1
        self.assertEqual(sorted(orig_authors),
                         sorted(self.branch1.authors))

    def test_merge_with_bugs(self):
        '''A merge from a branch with authors'''
        bugs = ['https://launchpad.net/bugs/1 fixed',
                'https://bugzilla.gnome.org/show_bug.cgi?id=1 fixed',
                'https://launchpad.net/bugs/2 fixed']
        revprops = {}
        revprops['bugs'] = '\n'.join(bugs)
        self.branch1.commit('Bugs test', revprops)
        self.branch1.commit('Testing')
        self.branch2.commit(
            'Foo', revprops={'bugs': 'https://launchpad.net/bugs/3 fixed'})
        self.branch2.lp_branch.revision_count += 1
        self.branch2.merge(self.branch1)
        self.branch2.commit('Landed bugs')
        self.assertEqual(self.branch2.fixed_bugs, self.branch1.fixed_bugs)

    def test_merge_with_reviews(self):
        '''A merge with reviewers.'''
        reviews = ['reviewer1 Approve', 'reviewer2 Abstain']
        self.branch1.merge(self.branch2)
        self.branch1.commit('Reviewers test', reviews=reviews)

        last_rev = self.branch1.lp_branch._internal_bzr_branch.last_revision()
        self.assertNotEquals(last_rev, 'null:')
        repo = self.branch1.lp_branch._internal_bzr_branch.repository
        rev = repo.get_revision(last_rev)
        self.assertEquals('\n'.join(reviews),
                          rev.properties.get('reviews', None))

    def test_cleanup(self):
        '''The branch object should clean up after itself.'''
        self.branch1.merge(self.branch2)
        self.assertTrue(self.branch1.tree.changes_from(
                self.branch1.tree.basis_tree()).has_changed())
        with open(os.path.join(self.branch1.config.tree_dir, 'README.bak~'),
                  'w') as f:
            f.close()

        self.branch1.cleanup()
        self.assertFalse(self.branch1.tree.changes_from(
                self.branch1.tree.basis_tree()).has_changed())

    def test_unmanaged_files(self):
        """Test that the unmanaged_files property returns correct lists."""
        self.branch1.merge(self.branch2)
        self.assertEqual(self.branch1.unmanaged_files, [])
        expected = sorted([u'README~', u'newfile'])
        with open(os.path.join(self.branch1.config.tree_dir, 'README~'),
                  'w') as f:
            f.close()
        with open(os.path.join(self.branch1.config.tree_dir, 'newfile'),
                  'w') as f:
            f.close()
        self.assertEqual(sorted(self.branch1.unmanaged_files), expected)
        self.branch1.cleanup()
        self.assertEqual(self.branch1.unmanaged_files, [])

    def test_get_merge_proposals(self):
        """Test that the branch can find its own merge proposals."""
        self.assertEqual(self.branch1.get_merge_proposals(), [])
        proposals = [Thing(
            queue_status="Approved",
            commit_message="Goo goo g'joob"),]
        branch_with_proposals = branch.Branch.create(
            self.make_branch(proposals=proposals), self.config)
        self.assertEqual(
            branch_with_proposals.get_merge_proposals(), proposals)

    def test_get_reviews(self):
        """Test that the _get_reviews method gives the right lists."""
        self.assertEqual(self.branch1._get_reviews(self.proposals[0]),
                         [u'Reviewer;Needs Fixing'])
        self.assertEqual(self.branch1._get_reviews(self.proposals[1]),
                         [u'Reviewer;Approve', u'Reviewer2;Abstain'])

    def test_merge_branches(self):
        """Test that the merge_branches method merges approved branches."""
        mock = self.make_branch()
        proposals = [Thing(
            self_link=u'http://api.edge.launchpad.net/devel/proposal1',
            source_branch=self.branch2.lp_branch,
            target_branch=mock,
            queue_status="Approved",
            commit_message="Goo goo g'joob",
            prerequisite_branch=None,
            createComment=self.createComment,
            setStatus=self.lp_save,
            lp_save=self.lp_save,
            reviewed_revid=self.branch2.bzr_branch.last_revision(),
            votes=[Thing(
                    comment=Thing(vote=u'Approve'),
                    reviewer=Thing(display_name=u'Reviewer')),
                                   Thing(
                    comment=Thing(vote=u'Abstain'),
                    reviewer=Thing(display_name=u'Reviewer2'))])]
        mock.landing_candidates = proposals
        branch_with_proposals = branch.Branch.create(mock, self.config)
        results = branch_with_proposals.merge_branches()
        # Since we have only one proposal, test shows merge_branches() yields
        # control back to the caller twice, once for the pre commit hook and
        # once for the post commit hook
        (hook, source, proposal) = results.next()
        self.assertEqual(hook, "tarmac_pre_commit")
        self.assertEqual(source.lp_branch, self.branch2.lp_branch)
        self.assertEqual(proposal, proposals[0])
        (hook, source, proposal) = results.next()
        self.assertEqual(hook, "tarmac_post_commit")
        self.assertEqual(source.lp_branch, self.branch2.lp_branch)
        self.assertEqual(proposal, proposals[0])
        # Calling results.next on the exhausted generator raises a
        # StopIteration
        self.assertRaises(StopIteration, results.next)
        # Show self.branch2 has been committed into branch_with_proposals.
        bzr_branch = branch_with_proposals.bzr_branch
        last_rev = bzr_branch.last_revision()
        rev = bzr_branch.repository.get_revision(last_rev)
        # Everything in the proposal is now in the last committed revision.
        self.assertEqual(rev.get_summary(), "Goo goo g'joob")
        self.assertEqual(rev.committer, "Tarmac")
        self.assertEqual(
            rev.properties.get('merge_url'),
            'http://code.launchpad.net/proposal1')
